"""Token-related utilities"""

# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from collections import namedtuple
from io import StringIO
from keyword import iskeyword

import tokenize
from tokenize import TokenInfo
from typing import List, Optional


Token = namedtuple('Token', ['token', 'text', 'start', 'end', 'line'])

def generate_tokens(readline):
    """wrap generate_tkens to catch EOF errors"""
    try:
        for token in tokenize.generate_tokens(readline):
            yield token
    except tokenize.TokenError:
        # catch EOF error
        return


def generate_tokens_catch_errors(
    readline, extra_errors_to_catch: Optional[List[str]] = None
):
    default_errors_to_catch = [
        "unterminated string literal",
        "invalid non-printable character",
        "after line continuation character",
    ]
    assert extra_errors_to_catch is None or isinstance(extra_errors_to_catch, list)
    errors_to_catch = default_errors_to_catch + (extra_errors_to_catch or [])

    tokens: List[TokenInfo] = []
    try:
        for token in tokenize.generate_tokens(readline):
            tokens.append(token)
            yield token
    except tokenize.TokenError as exc:
        if any(error in exc.args[0] for error in errors_to_catch):
            if tokens:
                start = tokens[-1].start[0], tokens[-1].end[0]
                end = start
                line = tokens[-1].line
            else:
                start = end = (1, 0)
                line = ""
            yield tokenize.TokenInfo(tokenize.ERRORTOKEN, "", start, end, line)
        else:
            # Catch EOF
            raise


def line_at_cursor(cell, cursor_pos=0):
    """Return the line in a cell at a given cursor position

    Used for calling line-based APIs that don't support multi-line input, yet.

    Parameters
    ----------
    cell : str
        multiline block of text
    cursor_pos : integer
        the cursor position

    Returns
    -------
    (line, offset): (string, integer)
        The line with the current cursor, and the character offset of the start of the line.
    """
    offset = 0
    lines = cell.splitlines(True)
    for line in lines:
        next_offset = offset + len(line)
        if not line.endswith('\n'):
            # If the last line doesn't have a trailing newline, treat it as if
            # it does so that the cursor at the end of the line still counts
            # as being on that line.
            next_offset += 1
        if next_offset > cursor_pos:
            break
        offset = next_offset
    else:
        line = ""
    return (line, offset)


def token_at_cursor(cell: str, cursor_pos: int = 0):
    """Get the token at a given cursor

    Used for introspection.

    Function calls are prioritized, so the token for the callable will be returned
    if the cursor is anywhere inside the call.

    Parameters
    ----------
    cell : str
        A block of Python code
    cursor_pos : int
        The location of the cursor in the block where the token should be found
    """
    names: List[str] = []
    tokens: List[Token] = []
    call_names = []
    
    offsets = {1: 0} # lines start at 1
    for tup in generate_tokens(StringIO(cell).readline):
        
        tok = Token(*tup)
        
        # token, text, start, end, line = tup
        start_line, start_col = tok.start
        end_line, end_col = tok.end
        if end_line + 1 not in offsets:
            # keep track of offsets for each line
            lines = tok.line.splitlines(True)
            for lineno, line in enumerate(lines, start_line + 1):
                if lineno not in offsets:
                    offsets[lineno] = offsets[lineno-1] + len(line)
        
        offset = offsets[start_line]
        # allow '|foo' to find 'foo' at the beginning of a line
        boundary = cursor_pos + 1 if start_col == 0 else cursor_pos
        if offset + start_col >= boundary:
            # current token starts after the cursor,
            # don't consume it
            break
        
        if tok.token == tokenize.NAME and not iskeyword(tok.text):
            if names and tokens and tokens[-1].token == tokenize.OP and tokens[-1].text == '.':
                names[-1] = "%s.%s" % (names[-1], tok.text)
            else:
                names.append(tok.text)
        elif tok.token == tokenize.OP:
            if tok.text == '=' and names:
                # don't inspect the lhs of an assignment
                names.pop(-1)
            if tok.text == '(' and names:
                # if we are inside a function call, inspect the function
                call_names.append(names[-1])
            elif tok.text == ')' and call_names:
                call_names.pop(-1)
        
        tokens.append(tok)
        
        if offsets[end_line] + end_col > cursor_pos:
            # we found the cursor, stop reading
            break
        
    if call_names:
        return call_names[-1]
    elif names:
        return names[-1]
    else:
        return ''



"""
Tests for platutils.py
"""

#-----------------------------------------------------------------------------
#  Copyright (C) 2008-2011  The IPython Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import sys
import signal
import os
import time
from _thread import interrupt_main  # Py 3
import threading

import pytest

from IPython.utils.process import (find_cmd, FindCmdError, arg_split,
                                   system, getoutput, getoutputerror,
                                   get_output_error_code)
from IPython.utils.capture import capture_output
from IPython.testing import decorators as dec
from IPython.testing import tools as tt

python = os.path.basename(sys.executable)

#-----------------------------------------------------------------------------
# Tests
#-----------------------------------------------------------------------------


@dec.skip_win32
def test_find_cmd_ls():
    """Make sure we can find the full path to ls."""
    path = find_cmd("ls")
    assert path.endswith("ls")

    
@dec.skip_if_not_win32
def test_find_cmd_pythonw():
    """Try to find pythonw on Windows."""
    path = find_cmd('pythonw')
    assert path.lower().endswith('pythonw.exe'), path


def test_find_cmd_fail():
    """Make sure that FindCmdError is raised if we can't find the cmd."""
    pytest.raises(FindCmdError, find_cmd, "asdfasdf")


@dec.skip_win32
@pytest.mark.parametrize(
    "argstr, argv",
    [
        ("hi", ["hi"]),
        ("hello there", ["hello", "there"]),
        # \u01ce == \N{LATIN SMALL LETTER A WITH CARON}
        # Do not use \N because the tests crash with syntax error in
        # some cases, for example windows python2.6.
        ("h\u01cello", ["h\u01cello"]),
        ('something "with quotes"', ["something", '"with quotes"']),
    ],
)
def test_arg_split(argstr, argv):
    """Ensure that argument lines are correctly split like in a shell."""
    assert arg_split(argstr) == argv


@dec.skip_if_not_win32
@pytest.mark.parametrize(
    "argstr,argv",
    [
        ("hi", ["hi"]),
        ("hello there", ["hello", "there"]),
        ("h\u01cello", ["h\u01cello"]),
        ('something "with quotes"', ["something", "with quotes"]),
    ],
)
def test_arg_split_win32(argstr, argv):
    """Ensure that argument lines are correctly split like in a shell."""
    assert arg_split(argstr) == argv


class SubProcessTestCase(tt.TempFileMixin):
    def setUp(self):
        """Make a valid python temp file."""
        lines = [ "import sys",
                 "print('on stdout', end='', file=sys.stdout)",
                 "print('on stderr', end='', file=sys.stderr)",
                 "sys.stdout.flush()",
                 "sys.stderr.flush()"]
        self.mktmp('\n'.join(lines))

    def test_system(self):
        status = system(f'{python} "{self.fname}"')
        self.assertEqual(status, 0)

    def test_system_quotes(self):
        status = system('%s -c "import sys"' % python)
        self.assertEqual(status, 0)

    def assert_interrupts(self, command):
        """
        Interrupt a subprocess after a second.
        """
        if threading.main_thread() != threading.current_thread():
            raise pytest.skip("Can't run this test if not in main thread.")

        # Some tests can overwrite SIGINT handler (by using pdb for example),
        # which then breaks this test, so just make sure it's operating
        # normally.
        signal.signal(signal.SIGINT, signal.default_int_handler)

        def interrupt():
            # Wait for subprocess to start:
            time.sleep(0.5)
            interrupt_main()

        threading.Thread(target=interrupt).start()
        start = time.time()
        try:
            result = command()
        except KeyboardInterrupt:
            # Success!
            pass
        end = time.time()
        self.assertTrue(
            end - start < 2, "Process didn't die quickly: %s" % (end - start)
        )
        return result

    def test_system_interrupt(self):
        """
        When interrupted in the way ipykernel interrupts IPython, the
        subprocess is interrupted.
        """
        def command():
            return system('%s -c "import time; time.sleep(5)"' % python)

        status = self.assert_interrupts(command)
        self.assertNotEqual(
            status, 0, f"The process wasn't interrupted. Status: {status}"
        )

    def test_getoutput(self):
        out = getoutput(f'{python} "{self.fname}"')
        # we can't rely on the order the line buffered streams are flushed
        try:
            self.assertEqual(out, 'on stderron stdout')
        except AssertionError:
            self.assertEqual(out, 'on stdouton stderr')

    def test_getoutput_quoted(self):
        out = getoutput('%s -c "print (1)"' % python)
        self.assertEqual(out.strip(), '1')

    #Invalid quoting on windows
    @dec.skip_win32
    def test_getoutput_quoted2(self):
        out = getoutput("%s -c 'print (1)'" % python)
        self.assertEqual(out.strip(), '1')
        out = getoutput("%s -c 'print (\"1\")'" % python)
        self.assertEqual(out.strip(), '1')

    def test_getoutput_error(self):
        out, err = getoutputerror(f'{python} "{self.fname}"')
        self.assertEqual(out, 'on stdout')
        self.assertEqual(err, 'on stderr')

    def test_get_output_error_code(self):
        quiet_exit = '%s -c "import sys; sys.exit(1)"' % python
        out, err, code = get_output_error_code(quiet_exit)
        self.assertEqual(out, '')
        self.assertEqual(err, '')
        self.assertEqual(code, 1)
        out, err, code = get_output_error_code(f'{python} "{self.fname}"')
        self.assertEqual(out, 'on stdout')
        self.assertEqual(err, 'on stderr')
        self.assertEqual(code, 0)

        



#  tests for ipython-8.20.0-py312hd43f75c_0 (this is a generated file);
print('===== testing package: ipython-8.20.0-py312hd43f75c_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import subprocess
import platform
import os
import sys
import IPython
import IPython.core
import IPython.core.magics
import IPython.core.tests
import IPython.extensions
import IPython.extensions.tests
import IPython.external
import IPython.lib
import IPython.lib.tests
import IPython.sphinxext
import IPython.terminal
import IPython.terminal.pt_inputhooks
import IPython.terminal.tests
import IPython.testing
import IPython.testing.plugin
import IPython.testing.tests
import IPython.utils.tests
import IPython.utils

WIN = platform.system() == "Windows"
LINUX = platform.system() == "Linux"
PYPY = "__pypy__" in sys.builtin_module_names
PPC = "ppc" in platform.machine()

COV_THRESHOLD = os.environ.get("COV_THRESHOLD")

# Environment variable should be set in the meta.yaml
MIGRATING = eval(os.environ.get("MIGRATING", "None"))

PYTEST_SKIPS = ["decorator_skip", "pprint_heap_allocated"]
PYTEST_ARGS = [sys.executable, "-m", "pytest", "--pyargs", "IPython", "-vv"]

if WIN:
    pass
else:
    pass

if LINUX:
    PYTEST_SKIPS += ["system_interrupt"]

if PPC:
    PYTEST_SKIPS += ["ipython_dir_8", "audio_data", "figure_to_svg", "figure_to_jpeg", "retina_figure", "select_figure_formats_kwargs", "test_debug_magic_passes_through_generators"]

if COV_THRESHOLD is not None:
    PYTEST_ARGS += [
        "--cov", "IPython", "--no-cov-on-fail", "--cov-fail-under", COV_THRESHOLD,
        "--cov-report", "term-missing:skip-covered"
    ]

if len(PYTEST_SKIPS) == 1:
    PYTEST_ARGS += ["-k", f"not {PYTEST_SKIPS[0]}"]
elif PYTEST_SKIPS:
    PYTEST_ARGS += ["-k", f"""not ({" or ".join(PYTEST_SKIPS) })"""]

if __name__ == "__main__":
    print("Building on Windows?", WIN)
    print("Building on Linux?  ", LINUX)
    print("Building for PyPy?  ", PYPY)

    if MIGRATING:
        print("This is a migration, skipping test suite! Put it back later!", flush=True)
        sys.exit(0)
    else:
        print("Running pytest with args")
        print(PYTEST_ARGS, flush=True)
        sys.exit(subprocess.call(PYTEST_ARGS))
#  --- run_test.py (end) ---

print('===== ipython-8.20.0-py312hd43f75c_0 OK =====');



from typing import List

import pytest
import pygments.lexers
import pygments.lexer

from IPython.lib.lexers import IPythonConsoleLexer, IPythonLexer, IPython3Lexer

#: the human-readable names of the IPython lexers with ``entry_points``
EXPECTED_LEXER_NAMES = [
    cls.name for cls in [IPythonConsoleLexer, IPythonLexer, IPython3Lexer]
]


@pytest.fixture
def all_pygments_lexer_names() -> List[str]:
    """Get all lexer names registered in pygments."""
    return {l[0] for l in pygments.lexers.get_all_lexers()}


@pytest.mark.parametrize("expected_lexer", EXPECTED_LEXER_NAMES)
def test_pygments_entry_points(
    expected_lexer: str, all_pygments_lexer_names: List[str]
) -> None:
    """Check whether the ``entry_points`` for ``pygments.lexers`` are correct."""
    assert expected_lexer in all_pygments_lexer_names



"""Utility functions for finding modules

Utility functions for finding modules on sys.path.

"""
#-----------------------------------------------------------------------------
# Copyright (c) 2011, the IPython Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

# Stdlib imports
import importlib
import sys

# Third-party imports

# Our own imports


#-----------------------------------------------------------------------------
# Globals and constants
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Local utilities
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Classes and functions
#-----------------------------------------------------------------------------

def find_mod(module_name):
    """
    Find module `module_name` on sys.path, and return the path to module `module_name`.

    * If `module_name` refers to a module directory, then return path to `__init__` file.
        * If `module_name` is a directory without an __init__file, return None.

    * If module is missing or does not have a `.py` or `.pyw` extension, return None.
        * Note that we are not interested in running bytecode.

    * Otherwise, return the fill path of the module.

    Parameters
    ----------
    module_name : str

    Returns
    -------
    module_path : str
        Path to module `module_name`, its __init__.py, or None,
        depending on above conditions.
    """
    spec = importlib.util.find_spec(module_name)
    module_path = spec.origin
    if module_path is None:
        if spec.loader in sys.meta_path:
            return spec.loader
        return None
    else:
        split_path = module_path.split(".")
        if split_path[-1] in ["py", "pyw"]:
            return module_path
        else:
            return None



# encoding: utf-8
"""Tests for IPython.utils.text"""

#-----------------------------------------------------------------------------
#  Copyright (C) 2011  The IPython Development Team
#
#  Distributed under the terms of the BSD License.  The full license is in
#  the file COPYING, distributed as part of this software.
#-----------------------------------------------------------------------------

#-----------------------------------------------------------------------------
# Imports
#-----------------------------------------------------------------------------

import os
import math
import random

from pathlib import Path

import pytest

from IPython.utils import text

#-----------------------------------------------------------------------------
# Globals
#-----------------------------------------------------------------------------


@pytest.mark.parametrize(
    "expected, width, row_first, spread",
    (
        (
            "aaaaa  bbbbb  ccccc  ddddd\n",
            80,
            False,
            False,
        ),
        (
            "aaaaa  ccccc\nbbbbb  ddddd\n",
            25,
            False,
            False,
        ),
        (
            "aaaaa  ccccc\nbbbbb  ddddd\n",
            12,
            False,
            False,
        ),
        (
            "aaaaa\nbbbbb\nccccc\nddddd\n",
            10,
            False,
            False,
        ),
        (
            "aaaaa  bbbbb  ccccc  ddddd\n",
            80,
            True,
            False,
        ),
        (
            "aaaaa  bbbbb\nccccc  ddddd\n",
            25,
            True,
            False,
        ),
        (
            "aaaaa  bbbbb\nccccc  ddddd\n",
            12,
            True,
            False,
        ),
        (
            "aaaaa\nbbbbb\nccccc\nddddd\n",
            10,
            True,
            False,
        ),
        (
            "aaaaa      bbbbb      ccccc      ddddd\n",
            40,
            False,
            True,
        ),
        (
            "aaaaa          ccccc\nbbbbb          ddddd\n",
            20,
            False,
            True,
        ),
        (
            "aaaaa  ccccc\nbbbbb  ddddd\n",
            12,
            False,
            True,
        ),
        (
            "aaaaa\nbbbbb\nccccc\nddddd\n",
            10,
            False,
            True,
        ),
    ),
)
def test_columnize(expected, width, row_first, spread):
    """Basic columnize tests."""
    size = 5
    items = [l*size for l in 'abcd']
    with pytest.warns(PendingDeprecationWarning):
        out = text.columnize(
            items, displaywidth=width, row_first=row_first, spread=spread
        )
        assert out == expected


def test_columnize_random():
    """Test with random input to hopefully catch edge case """
    for row_first in [True, False]:
        for nitems in [random.randint(2,70) for i in range(2,20)]:
            displaywidth = random.randint(20,200)
            rand_len = [random.randint(2,displaywidth) for i in range(nitems)]
            items = ['x'*l for l in rand_len]
            with pytest.warns(PendingDeprecationWarning):
                out = text.columnize(
                    items, row_first=row_first, displaywidth=displaywidth
                )
            longer_line = max([len(x) for x in out.split("\n")])
            longer_element = max(rand_len)
            assert longer_line <= displaywidth, (
                f"Columnize displayed something lager than displaywidth : {longer_line}\n"
                f"longer element : {longer_element}\n"
                f"displaywidth : {displaywidth}\n"
                f"number of element : {nitems}\n"
                f"size of each element : {rand_len}\n"
                f"row_first={row_first}\n"
            )


@pytest.mark.parametrize("row_first", [True, False])
def test_columnize_medium(row_first):
    """Test with inputs than shouldn't be wider than 80"""
    size = 40
    items = [l*size for l in 'abc']
    with pytest.warns(PendingDeprecationWarning):
        out = text.columnize(items, row_first=row_first, displaywidth=80)
    assert out == "\n".join(items + [""]), "row_first={0}".format(row_first)


@pytest.mark.parametrize("row_first", [True, False])
def test_columnize_long(row_first):
    """Test columnize with inputs longer than the display window"""
    size = 11
    items = [l*size for l in 'abc']
    with pytest.warns(PendingDeprecationWarning):
        out = text.columnize(items, row_first=row_first, displaywidth=size - 1)
    assert out == "\n".join(items + [""]), "row_first={0}".format(row_first)


def eval_formatter_check(f):
    ns = dict(n=12, pi=math.pi, stuff='hello there', os=os, u=u"café", b="café")
    s = f.format("{n} {n//4} {stuff.split()[0]}", **ns)
    assert s == "12 3 hello"
    s = f.format(" ".join(["{n//%i}" % i for i in range(1, 8)]), **ns)
    assert s == "12 6 4 3 2 2 1"
    s = f.format("{[n//i for i in range(1,8)]}", **ns)
    assert s == "[12, 6, 4, 3, 2, 2, 1]"
    s = f.format("{stuff!s}", **ns)
    assert s == ns["stuff"]
    s = f.format("{stuff!r}", **ns)
    assert s == repr(ns["stuff"])

    # Check with unicode:
    s = f.format("{u}", **ns)
    assert s == ns["u"]
    # This decodes in a platform dependent manner, but it shouldn't error out
    s = f.format("{b}", **ns)

    pytest.raises(NameError, f.format, "{dne}", **ns)


def eval_formatter_slicing_check(f):
    ns = dict(n=12, pi=math.pi, stuff='hello there', os=os)
    s = f.format(" {stuff.split()[:]} ", **ns)
    assert s == " ['hello', 'there'] "
    s = f.format(" {stuff.split()[::-1]} ", **ns)
    assert s == " ['there', 'hello'] "
    s = f.format("{stuff[::2]}", **ns)
    assert s == ns["stuff"][::2]

    pytest.raises(SyntaxError, f.format, "{n:x}", **ns)

def eval_formatter_no_slicing_check(f):
    ns = dict(n=12, pi=math.pi, stuff="hello there", os=os)

    s = f.format("{n:x} {pi**2:+f}", **ns)
    assert s == "c +9.869604"

    s = f.format("{stuff[slice(1,4)]}", **ns)
    assert s == "ell"

    s = f.format("{a[:]}", a=[1, 2])
    assert s == "[1, 2]"

def test_eval_formatter():
    f = text.EvalFormatter()
    eval_formatter_check(f)
    eval_formatter_no_slicing_check(f)

def test_full_eval_formatter():
    f = text.FullEvalFormatter()
    eval_formatter_check(f)
    eval_formatter_slicing_check(f)

def test_dollar_formatter():
    f = text.DollarFormatter()
    eval_formatter_check(f)
    eval_formatter_slicing_check(f)
    
    ns = dict(n=12, pi=math.pi, stuff='hello there', os=os)
    s = f.format("$n", **ns)
    assert s == "12"
    s = f.format("$n.real", **ns)
    assert s == "12"
    s = f.format("$n/{stuff[:5]}", **ns)
    assert s == "12/hello"
    s = f.format("$n $$HOME", **ns)
    assert s == "12 $HOME"
    s = f.format("${foo}", foo="HOME")
    assert s == "$HOME"


def test_strip_email():
    src = """\
        >> >>> def f(x):
        >> ...   return x+1
        >> ... 
        >> >>> zz = f(2.5)"""
    cln = """\
>>> def f(x):
...   return x+1
... 
>>> zz = f(2.5)"""
    assert text.strip_email_quotes(src) == cln


def test_strip_email2():
    src = "> > > list()"
    cln = "list()"
    assert text.strip_email_quotes(src) == cln


def test_LSString():
    lss = text.LSString("abc\ndef")
    assert lss.l == ["abc", "def"]
    assert lss.s == "abc def"
    lss = text.LSString(os.getcwd())
    assert isinstance(lss.p[0], Path)


def test_SList():
    sl = text.SList(["a 11", "b 1", "a 2"])
    assert sl.n == "a 11\nb 1\na 2"
    assert sl.s == "a 11 b 1 a 2"
    assert sl.grep(lambda x: x.startswith("a")) == text.SList(["a 11", "a 2"])
    assert sl.fields(0) == text.SList(["a", "b", "a"])
    assert sl.sort(field=1, nums=True) == text.SList(["b 1", "a 2", "a 11"])
